import sys
from types import ClassType, NoneType
from mext.classhacks import on_class_ready, is_in_class_declaration
from mext.dictbase import DictBase
from mext.lazyimport import onimport_cls
from mext.const import const


__all__ = [
    'DispatcherBase', 'TypeDispatcher',
    'GenericFunction', 'Adapter',
    'ProtocolBase', 'Interface'
]

# TODO: Adapter chains

NO_ARGUMENT = const()

def noop(ob):
    return ob


class DispatcherBase(DictBase):
    __slots__ = ('map', 'cache')

    def __init__(self):
        self.map = {}
        self.cache = {}

    def implies(self, a, b):
        raise NotImplementedError

    validate_key = validate_value = staticmethod(noop)

    def keys(self):
        return self.map.keys()

    def __setitem__(self, key, value):
        key = self.validate_key(key)
        if key in self.map:
            raise KeyError("Conflict for key", key, self.map[key])
        value = self.validate_value(value)
        self.map[key] = value
        self.cache = {}

    def __getitem__(self, key):
        values = self.query(key)
        if len(values) == 1:
            return values[0]
        elif not values:
            raise KeyError("Nothing registered for", key)
        else:
            topkeys = self._find_topkeys(key)
            raise LookupError("Can't dispatch", key, list(topkeys))

    def query(self, key):
        if key in self.cache:
            return self.cache[key]
        topkeys = self._find_topkeys(key)
        r = map(self.map.__getitem__, topkeys)
        if len(r) > 1:
            r = set(r)
        r = self.cache[key] = tuple(r)
        return r

    def _find_topkeys(self, req):
        implies = self.implies
        req = self.validate_key(req)
        if req in self.map:
            return [req]
        topkeys = set()
        for key in self.map:
            if implies(req, key):
                for topkey in list(topkeys):
                    if implies(topkey, key):
                        break
                    elif implies(key, topkey):
                        topkeys.remove(topkey)
                else:
                    topkeys.add(key)
        return topkeys



    def register(self, key, method=NO_ARGUMENT):
        if method is NO_ARGUMENT:
            return lambda m: self.register(key, m)
        self._register(key, method)
        return method

    def _register(self, key, method):
        self[key] = method


## class CompareDispatcher(DispatcherBase):
##     implies = operator.le
##     """
##     Here `a <= b` means a is less generic than b
##     """
##



class TypeDispatcher(DispatcherBase):
    __slots__ = ()
    def implies(self, a, b):
        if b is object:
            return True
        else:
            return issubclass(a, b)

    def validate_key(self, key):
        if key is None:
            return NoneType
        elif isinstance(key, (type, ClassType)):
            return key
        else:
            raise TypeError("TypeDispatcher only works on classes, got %r" % type(key))


    def register_for_this_class(self, method, depth=0):
        on_class_ready(lambda cls: self.register(cls, method), depth=depth+1)
        return method

    for_this_class = register_for_this_class

    def register_this_class_for(self, key, depth=0):
        on_class_ready(self.register(key), depth=depth+1)

    def _register(self, cls, method):
        register = super(TypeDispatcher, self)._register
        if isinstance(cls, str):
            @onimport_cls(cls)
            def onimport(cls):
                register(cls, method)
        else:
            register(cls, method)


class GenericFunction(TypeDispatcher):
    __slots__ = ('method_name',)
    def __init__(self, method_name=None):
        self.method_name = method_name
        TypeDispatcher.__init__(self)

    def validate_value(self, value):
        if not callable(value):
            raise TypeError("Generic function items have to be callable", value)
        return value


    def __call__(self, ob, *args):
        mname = self.method_name
        if mname is not None:
            try:
                method = getattr(ob, mname)
            except AttributeError:
                pass
            else:
                return method(*args)
        try:
            method = self[ob.__class__]
        except KeyError:
            raise TypeError("No methods for %s" % ob)
        else:
            return method(ob, *args)


#@@ remove subadapt argument, use ClassAdapter subclass

class Adapter(GenericFunction):
    def __init__(self, method_name=None, subadapt=True):
        GenericFunction.__init__(self, method_name)
        self.subadapt = subadapt
        if is_in_class_declaration(1): # we do need this at least because subadapt is optional
            self._register_noop_for_enclosing_class(1)

    def _register_noop_for_enclosing_class(self, depth):
        on_class_ready(self.register_noop, depth=depth+1)

    def register_noop(self, cls=None, depth=0):
        if cls is None:
            self._register_noop_for_enclosing_class(depth+1)
        elif isinstance(cls, str):
            onimport_cls(cls, self.register_noop)
        else:
            self[cls] = noop
            #return cls # this enables us to use .register_noop as a class decorator

    def __get__(self, ob, type):
        if self.subadapt:
            return SubAdapter(self, type)
        else:
            return self

    def property(self, default=None):
        return AdaptingProperty(self, default)






class AdaptingProperty(object):
    def __init__(self, adapter, default):
        self.adapter = adapter
        self.default = default

    def __get__(self, ob, type):
        if ob is None:
            return self
        return ob.__dict__.get(self, self.default)

    def __set__(self, ob, value):
        ob.__dict__[self] = self.adapter(value)


class proxy_method(object):
    def __init__(self, name):
        self.name = name

    def __get__(self, ob, type):
        if ob is None:
            return self
        return getattr(ob.adapt, self.name)

class AdapterProxy(object):
    for name in [
        '__setitem__',
        'register',
        'register_noop',
        'register_for_this_class',
        'register_this_class_for'
    ]:
        locals()[name] = proxy_method(name)
    del name



class SubAdapter(AdapterProxy):
    def __init__(self, adapt, cls):
        self.adapt = adapt
        self.cls = cls

    def __call__(self, ob):
        if not isinstance(ob, self.cls):
            ob = self.adapt(ob)
            if not isinstance(ob, self.cls):
                raise TypeError("Adaptation result %r is not an instance of %s" % (ob, self.cls))
        return ob




class AdaptingType(object):
    __slots__ = ()
    class __metaclass__(type):
        def __call__(self, ob):
            return self.adapt(ob)

        def __setitem__(self, key, value):
            self.adapt[key] = value

    @classmethod
    def property(cls, default=None):
        return AdaptingProperty(cls.adapt, default)


class ProtocolBase(AdaptingType):
    """
        Overrides subclasses' constructor to try to adapt to the class by
        calling cls.adapt(ob) on constructor argument.

        Pure constructor (__init__ as defined in subclass) is exposed
        as cls.new(ob), so if one were to declare that for `str` argument
        to constructor no special adaptation were needed and it should be
        passed to normal constructor, one would write:
            cls[str] = cls.new
        or
            cls.adapt[str] = cls.new
    """
    __slots__ = ()
    @classmethod
    def new(*args, **kw):
        return type.__call__(*args, **kw)




class Interface(AdaptingType):
    """
        Baseclass for declarative protocol definitions.
        No instances of this class or subclasses are ever created.
    """
    class __metaclass__(AdapterProxy, AdaptingType.__metaclass__):
        def __init__(self, name, bases, dct):
            if bases == (AdaptingType,):
                # this is not a subclass, but mext.dispatch.Interface itself
                return
            if hasattr(self, 'adapt'):
                if 'adapt' not in dct:
                    raise TypeError("An Interface subclass is expected to redefine the adapter")
                assert not self.adapt.subadapt
                adapt = self.adapt
            else:
                adapt = Adapter(subadapt=False)
            self.adapt = adapt

        def __repr__(self):
            return "<interface '%s.%s'>" % (self.__module__, self.__name__)




